<?php
/**
 * Exception handler for the Horde_Ldap library.
 *
 * Copyright 2010-2017 Horde LLC (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://www.gnu.org/licenses/lgpl-3.0.html.
 *
 * @category Horde
 * @package  Ldap
 * @author   Ben Klang <ben@alkaloid.net>
 */
class Horde_Ldap_Exception extends Horde_Exception_Wrapped
{
}
